@@ -2,7 +2,9 @@ from .base import (
     ConvertResponse,
     CreateRegister,
     CreatorDecorator,
+    Creator,
     ConvertFunction,
+    ConvertDefinition,
     Registerer,
     Ann,
     Annotation,
@@ -10,6 +12,7 @@ from .base import (
     FromMeta,
     NotSpecified,
 )
+from .hints import resolve_types
 from .version import VERSION
 from .meta import Meta
 from . import errors
@@ -23,10 +26,13 @@ __all__ = [
     "FromMeta",
     "MergedAnnotation",
     "ConvertFunction",
+    "ConvertDefinition",
     "ConvertResponse",
     "CreateRegister",
     "CreatorDecorator",
+    "Creator",
     "NotSpecified",
     "Registerer",
+    "resolve_types",
     "errors",
 ]
