@@ -1,8 +1,8 @@
 from .meta import Meta, extract_type
+from .hints import resolve_types
 from . import errors
 
 from cattrs.errors import IterableValidationError
-from functools import partial
 from attrs import define
 import typing as tp
 import inspect
@@ -22,38 +22,27 @@ class NotSpecified(metaclass=NotSpecifiedMeta):
         raise Exception("Do not instantiate NotSpecified")
 
 
-ConvertResponseValues: tp.TypeAlias = tp.Optional[bool | dict[str, tp.Any]]
-ConvertResponseGenerator: tp.TypeAlias = tp.Generator[ConvertResponseValues | T, T, None]
-
-ConvertResponse: tp.TypeAlias = ConvertResponseValues | ConvertResponseGenerator[T] | T
-
-ConvertDefinitionValue: tp.TypeAlias = tp.Callable[[tp.Any], ConvertResponse[T]]
-ConvertDefinitionNoValue: tp.TypeAlias = tp.Callable[[], ConvertResponse[T]]
-ConvertDefinitionValueAndType: tp.TypeAlias = tp.Callable[[tp.Any, tp.Type], ConvertResponse[T]]
-
-
-class ConvertDefinitionValueAndData(tp.Protocol):
-    def __call__(self, value: tp.Any, /, *values: tp.Any) -> ConvertResponse[T]:
-        ...
-
-
-class ConvertDefinitionValueAndTypeAndData(tp.Protocol):
-    def __call__(self, value: tp.Any, want: tp.Type, /, *values: tp.Any) -> ConvertResponse[T]:
-        ...
-
-
-ConvertDefinition: tp.TypeAlias = tp.Union[
-    ConvertDefinitionValue[T],
-    ConvertDefinitionNoValue[T],
-    ConvertDefinitionValueAndType[T],
-    ConvertDefinitionValueAndData,
-    ConvertDefinitionValueAndTypeAndData,
+ConvertResponseValues: tp.TypeAlias = bool | dict[str, tp.Any] | T
+ConvertResponseGenerator: tp.TypeAlias = tp.Generator[
+    tp.Optional[ConvertResponseValues[T]], T, None
 ]
-ConvertFunction: tp.TypeAlias = tp.Callable[[tp.Any, tp.Type, Meta, cattrs.Converter], T]
+
+ConvertResponse: tp.TypeAlias = tp.Optional[ConvertResponseValues[T] | ConvertResponseGenerator[T]]
+
+ConvertDefinitionNoValue: tp.TypeAlias = tp.Callable[[], ConvertResponse[T]]
+ConvertDefinitionValue: tp.TypeAlias = tp.Callable[[tp.Any], ConvertResponse[T]]
+ConvertDefinitionValueAndType: tp.TypeAlias = tp.Callable[[tp.Any, tp.Type], ConvertResponse[T]]
+# Also allowed is
+# - (Any, Type, /, meta1, meta2, ...)
+# - (Any, /, meta1, meta2, ...)
+# But python typing is restrictive and you can't express that
+
+ConvertDefinition: tp.TypeAlias = tp.Callable[..., ConvertResponse[T]]
+ConvertFunction: tp.TypeAlias = tp.Callable[["CreateArgs"], T]
 
 
 class WrappedCreator(tp.Generic[T]):
-    def __init__(self, wrapped: tp.Callable[["CreateArgs"], T], func: ConvertFunction[T]):
+    def __init__(self, wrapped: tp.Callable[["CreateArgs"], T], func: ConvertDefinition[T]):
         self.func = func
         self.wrapped = wrapped
 
@@ -64,7 +53,7 @@ class WrappedCreator(tp.Generic[T]):
         return f"<Wrapped {self.func}>"
 
 
-def take_or_make(value: tp.Any, typ: tp.Type[T], /) -> ConvertResponse:
+def take_or_make(value: tp.Any, typ: tp.Type[T], /) -> ConvertResponse[T]:
     if isinstance(value, typ):
         return value
     elif value is NotSpecified or isinstance(value, dict):
@@ -74,7 +63,7 @@ def take_or_make(value: tp.Any, typ: tp.Type[T], /) -> ConvertResponse:
 
 
 def filldict(
-    converter: cattrs.Converter, register: "CreateRegister", res: tp.Any, want: T
+    converter: cattrs.Converter, register: "CreateRegister", res: tp.Any, want: tp.Type
 ) -> tp.Any:
     if isinstance(res, dict) and hasattr(want, "__attrs_attrs__"):
         for field in attrs.fields(want):
@@ -84,7 +73,9 @@ def filldict(
     return res
 
 
-def fromdict(converter: cattrs.Converter, register: "CreateRegister", res: tp.Any, want: T) -> T:
+def fromdict(
+    converter: cattrs.Converter, register: "CreateRegister", res: tp.Any, want: type[T]
+) -> T:
     if res is NotSpecified:
         res = {}
     res = filldict(converter, register, res, want)
@@ -106,58 +97,56 @@ class MergedAnnotation(Annotation):
 
 
 @tp.runtime_checkable
-class _Ann(tp.Protocol):
-    def adjusted_meta(self, meta: Meta, typ: tp.Type) -> Meta:
+class _Ann(tp.Protocol[T]):
+    def adjusted_meta(self, meta: Meta, typ: tp.Type[T]) -> Meta:
         ...
 
     def adjusted_creator(
-        self, creator: ConvertFunction, register: "CreateRegister", typ: tp.Type[T]
-    ) -> ConvertFunction[T]:
+        self, creator: tp.Optional[ConvertFunction[T]], register: "CreateRegister", typ: tp.Type[T]
+    ) -> tp.Optional[ConvertFunction[T]]:
         ...
 
-    def bypass(self) -> int:
-        return 0
-
-
-@define(frozen=True)
-class Bypass(_Ann):
-    amount: int = 1
-
-    def bypass(self) -> int:
-        return self.amount
-
 
 @define(frozen=True)
 class FromMeta(_Ann):
     pattern: str
 
-    def adjusted_meta(self, meta: Meta, typ: tp.Type) -> Meta:
+    def adjusted_meta(self, meta: Meta, typ: tp.Type[T]) -> Meta:
         val = meta.retrieve_one(typ, self.pattern)
         return meta.clone(data_override={"retrieved": val})
 
     def adjusted_creator(
-        self, creator: ConvertFunction, register: "CreateRegister", typ: tp.Type[T]
-    ) -> ConvertFunction[T]:
-        def retrieve(val: tp.Any, /, retrieved: typ) -> typ:
-            return retrieved
+        self, creator: tp.Optional[ConvertFunction], register: "CreateRegister", typ: tp.Type[T]
+    ) -> tp.Optional[ConvertFunction[T]]:
+        def retrieve(val: tp.Any, /, _meta: Meta) -> ConvertResponse[T]:
+            return tp.cast(T, _meta.retrieve_one(object, "retrieved"))
 
-        return Ann(creator=retrieve).adjusted_creator(creator, register, typ)
+        a = Ann[T](creator=retrieve)
+        return a.adjusted_creator(creator, register, typ)
 
 
-class Ann(_Ann):
-    _func: tp.Optional[ConvertFunction] = None
+@tp.runtime_checkable
+class AdjustableMeta(tp.Protocol):
+    def adjusted_meta(self, meta: Meta, typ: tp.Type[T]) -> Meta:
+        ...
+
+
+class Ann(_Ann[T]):
+    _func: tp.Optional[ConvertFunction[T]] = None
 
     def __init__(
-        self, meta: tp.Optional[Annotation] = None, creator: tp.Optional[ConvertFunction] = None
+        self,
+        meta: tp.Optional[Annotation | AdjustableMeta] = None,
+        creator: tp.Optional[ConvertDefinition[T]] = None,
     ):
         self.meta = meta
         self.creator = creator
 
-    def adjusted_meta(self, meta: Meta, typ: tp.Type) -> Meta:
+    def adjusted_meta(self, meta: Meta, typ: tp.Type[T]) -> Meta:
         if self.meta is None:
             return meta
 
-        if hasattr(self.meta, "adjusted_meta"):
+        if isinstance(self.meta, AdjustableMeta):
             return self.meta.adjusted_meta(meta, typ)
 
         if self.meta.merge_meta:
@@ -173,49 +162,80 @@ class Ann(_Ann):
             return meta.clone({"__call_defined_annotation__": self.meta})
 
     def adjusted_creator(
-        self, creator: ConvertFunction, register: "CreateRegister", typ: tp.Type[T]
-    ) -> ConvertFunction[T]:
+        self, creator: tp.Optional[ConvertFunction], register: "CreateRegister", typ: tp.Type[T]
+    ) -> tp.Optional[ConvertFunction[T]]:
         if self.creator is None:
             return creator
 
-        return CreatorDecorator(
-            register,
-            typ,
-            assume_unchanged_converted=hasattr(typ, "__attrs_attrs__"),
-            return_wrapped=True,
-        )(self.creator)
+        wrapped, _ = CreatorDecorator(
+            register, typ, assume_unchanged_converted=hasattr(typ, "__attrs_attrs__")
+        ).wrap(self.creator)
+
+        return wrapped
 
 
 class Registerer(tp.Protocol[T]):
-    def __init__(
-        self,
-        typ: tp.Type[T],
-        assume_unchanged_converted=True,
-    ):
-        ...
-
     def __call__(self, func: tp.Optional[ConvertDefinition[T]] = None) -> ConvertDefinition[T]:
         ...
 
 
+class Creator(tp.Protocol):
+    def __call__(self, typ: tp.Type[T], assume_unchanged_converted=True) -> Registerer[T]:
+        ...
+
+
 class CreateRegister:
-    def __init__(self):
-        self.register: dict[tp.Type[T], ConvertFunction[T]] = {}
+    def __init__(
+        self,
+        *,
+        register: None | dict[type[T], ConvertFunction[T]] = None,
+        last_meta: None | Meta = None,
+        last_type: None | type[T] = None,
+        skip_creator: None | ConvertDefinition[T] = None,
+        auto_resolve_attrs_and_dataclasses_annotations: bool = True,
+    ):
+        if register is None:
+            register = {}
+        self.register = register
+        self.last_meta = last_meta
+        self.last_type = last_type
+        self.skip_creator = skip_creator
+        self.auto_resolve_attrs_and_dataclasses_annotations = (
+            auto_resolve_attrs_and_dataclasses_annotations
+        )
+
+    def clone(
+        self, last_meta: Meta, last_type: type[T], skip_creator: ConvertDefinition[T]
+    ) -> "CreateRegister":
+        return type(self)(
+            register=self.register,
+            last_meta=last_meta,
+            last_type=last_type,
+            skip_creator=skip_creator,
+            auto_resolve_attrs_and_dataclasses_annotations=self.auto_resolve_attrs_and_dataclasses_annotations,
+        )
 
     def __setitem__(self, typ: tp.Type[T], creator: ConvertFunction[T]) -> None:
         if not isinstance(typ, type):
             raise errors.CanOnlyRegisterTypes(got=typ)
         self.register[typ] = creator
 
-    def __contains__(self, typ: tp.Type[T]) -> bool:
+    def __contains__(self, typ: tp.Type) -> bool:
         return self.creator_for(typ) is not None
 
-    def make_decorator(self) -> Registerer:
-        return partial(CreatorDecorator, self)
+    def make_decorator(self) -> Creator:
+        def creator(typ: tp.Type[T], assume_unchanged_converted=True) -> Registerer[T]:
+            return tp.cast(
+                Registerer[T],
+                CreatorDecorator(self, typ, assume_unchanged_converted=assume_unchanged_converted),
+            )
+
+        return creator
 
     def creator_for(self, typ: tp.Type[T]) -> tp.Optional[ConvertFunction[T]]:
-        if hasattr(typ, "__origin__"):
-            typ = typ.__origin__
+        origin: tp.Optional[tp.Type[T]] = getattr(typ, "__origin__", None)
+        if origin is not None:
+            typ = origin
 
         if not isinstance(typ, type):
             return None
@@ -235,10 +255,35 @@ class CreateRegister:
         value: tp.Any = NotSpecified,
         meta: tp.Any = NotSpecified,
         once_only_creator: tp.Optional[ConvertFunction[T]] = None,
-        recursed=False,
-    ):
+    ) -> T:
         return _CreateStructureHook.structure(
-            self, typ, value, meta, once_only_creator, recursed=recursed
+            self,
+            typ,
+            value,
+            meta,
+            once_only_creator,
+            last_meta=self.last_meta,
+            last_type=self.last_type,
+            skip_creator=self.skip_creator,
+        )
+
+    def create_annotated(
+        self,
+        typ: tp.Type[T],
+        ann: tp.Union[Annotation | _Ann | ConvertFunction[T]],
+        value: tp.Any = NotSpecified,
+        meta: tp.Any = NotSpecified,
+        once_only_creator: tp.Optional[ConvertFunction[T]] = None,
+    ) -> T:
+        return _CreateStructureHook.structure(
+            self,
+            tp.cast(tp.Type[T], tp.Annotated[typ, ann]),
+            value,
+            meta,
+            once_only_creator,
+            last_meta=self.last_meta,
+            last_type=self.last_type,
+            skip_creator=self.skip_creator,
         )
 
 
@@ -251,21 +296,33 @@ class _CreateStructureHook:
         value: tp.Any = NotSpecified,
         meta: tp.Any = NotSpecified,
         creator: tp.Optional[ConvertFunction[T]] = None,
-        recursed=False,
+        last_meta: None | Meta = None,
+        last_type: None | type[T] = None,
+        skip_creator: None | ConvertDefinition[T] = None,
     ) -> T:
         if meta is NotSpecified:
-            meta = Meta()
+            if last_meta is not None:
+                meta = last_meta.clone()
+            else:
+                meta = Meta()
 
         converter = meta.converter
-        hooks = kls(register, converter, meta, creator)
+        hooks = kls(
+            register=register,
+            converter=converter,
+            meta=meta,
+            last_meta=last_meta,
+            last_type=last_type,
+            skip_creator=skip_creator,
+            once_only_creator=creator,
+        )
+
+        ann, want = hooks._interpret_annotation(typ)
+
         converter.register_structure_hook_func(hooks.switch_check, hooks.convert)
 
         try:
-            if recursed:
-                # Skip this hook and the hook we recursed from
-                ret = hooks.bypass(value, tp.Annotated[typ, Bypass(2)])
-            else:
-                ret = hooks.convert(value, typ)
+            ret = hooks.convert(value, typ)
         finally:
             converter._structure_func._function_dispatch._handler_pairs.remove(
                 (hooks.switch_check, hooks.convert, False)
@@ -279,43 +336,41 @@ class _CreateStructureHook:
         converter: cattrs.Converter,
         meta: tp.Any = NotSpecified,
         once_only_creator: tp.Optional[ConvertFunction[T]] = None,
+        last_meta: None | Meta = None,
+        last_type: None | type[T] = None,
+        skip_creator: None | ConvertDefinition[T] = None,
     ):
         self.meta = meta
         self.register = register
         self.do_check = True
+        self.last_meta = last_meta
+        self.last_type = last_type
         self.converter = converter
+        self.skip_creator = skip_creator
         self.once_only_creator = once_only_creator
         self.cache: dict[tp.Type[T], ConvertFunction[T]] = {}
 
-    def _interpret_annotation(self, want: tp.Type) -> tp.Tuple["_Ann", tp.Type[T]]:
-        ann: tp.Optional[Annotation | _Ann | ConvertFunction] = None
-        if (
-            hasattr(want, "__metadata__")
-            and want.__metadata__
-            and (
-                isinstance(want.__metadata__[0], (_Ann, Annotation))
-                or callable(want.__metadata__[0])
-            )
-        ):
-            ann = want.__metadata__[0]
+    def _interpret_annotation(self, want: tp.Type[T]) -> tp.Tuple[tp.Optional[_Ann], tp.Type[T]]:
+        ann: tp.Optional[Annotation | _Ann | ConvertDefinition[T]] = None
+        if hasattr(want, "__metadata__"):
+            metadata: tp.Tuple[tp.Any] = getattr(want, "__metadata__")
+            if metadata and (isinstance(metadata[0], (_Ann, Annotation)) or callable(metadata[0])):
+                ann = metadata[0]
 
-            if isinstance(ann, Annotation):
-                ann = Ann(ann)
-            elif callable(ann):
-                ann = Ann(creator=ann)
+                if isinstance(ann, Annotation):
+                    ann = Ann(ann)
+                elif callable(ann):
+                    ann = Ann[T](creator=ann)
 
-            want = want.__origin__
+                origin = getattr(want, "__origin__", None)
+                if origin is not None:
+                    want = origin
 
-        return ann, want
+        return tp.cast(_Ann, ann), want
 
     def convert(self, value: tp.Any, want: tp.Type[T]) -> T:
+        wrapped: tp.Type[T]
         ann, wrapped = self._interpret_annotation(want)
-        if ann:
-            match bypass := ann.bypass():
-                case bypass if bypass > 1:
-                    return self.bypass(value, tp.Annotated[wrapped, Bypass(bypass - 1)])
-                case 1:
-                    return self.bypass(value, wrapped)
 
         if not self.check_func(want):
             return self.bypass(value, want)
@@ -328,25 +383,41 @@ class _CreateStructureHook:
         if want in self.cache and creator is None:
             creator = self.cache[want]
 
-        if ann:
+        if isinstance(ann, AdjustableMeta):
             meta = ann.adjusted_meta(meta, want)
             creator = ann.adjusted_creator(creator, self.register, want)
             if meta is not self.meta:
                 return self.register.create(want, value, meta=meta, once_only_creator=creator)
 
-        if creator:
-            return creator(CreateArgs(value, want, meta, self.converter, self.register))
-        elif isinstance(value, want):
+        if creator and creator is not self.skip_creator and getattr(creator, "func", None):
+            skip = False
+            if self.skip_creator:
+                if (
+                    creator is self.skip_creator
+                    or getattr(creator, "func", None) is self.skip_creator
+                ):
+                    skip = True
+
+            if not skip:
+                return creator(CreateArgs(value, want, meta, self.converter, self.register))
+
+        if isinstance(value, want):
             return value
         else:
             return fromdict(self.converter, self.register, value, want)
 
-    def switch_check(self, want: tp.Type[T]) -> bool:
+    def switch_check(self, want: tp.Type) -> bool:
+        if self.register.auto_resolve_attrs_and_dataclasses_annotations:
+            resolve_types(want)
+
         ret = self.do_check
         self.do_check = True
         return ret
 
-    def check_func(self, want: tp.Type[T]) -> bool:
+    def check_func(self, want: tp.Type) -> bool:
+        if self.register.auto_resolve_attrs_and_dataclasses_annotations:
+            resolve_types(want)
+
         ann, want = self._interpret_annotation(want)
         creator = self.register.creator_for(want)
 
@@ -354,7 +425,7 @@ class _CreateStructureHook:
             self.cache[want] = creator
             return True
 
-        return ann or self.once_only_creator or hasattr(want, "__attrs_attrs__")
+        return bool(ann or self.once_only_creator or hasattr(want, "__attrs_attrs__"))
 
     def bypass(self, value: tp.Any, want: tp.Type[T]) -> T:
         self.do_check = False
@@ -363,22 +434,25 @@ class _CreateStructureHook:
             value = filldict(self.converter, self.register, value, want)
             return self.converter.structure(value, want)
         except IterableValidationError as e:
-            raise errors.FailedToConvertIterable(message=e.message, exceptions=e.exceptions)
+            raise errors.FailedToConvertIterable(message=e.message, exceptions=list(e.exceptions))
 
 
 class _ArgsExtractor:
     def __init__(
         self,
+        *,
         signature: inspect.Signature,
         value: tp.Any,
         want: tp.Type,
         meta: Meta,
+        creator: ConvertDefinition,
         converter: cattrs.Converter,
         register: CreateRegister,
     ):
         self.meta = meta
         self.want = want
         self.value = value
+        self.creator = creator
         self.register = register
         self.converter = converter
         self.signature = signature
@@ -395,7 +469,7 @@ class _ArgsExtractor:
             values.pop(0)
             use.append(self.want)
 
-        def provided(param: inspect.Parameter, name: str, typ: tp.Type[T]) -> T:
+        def provided(param: inspect.Parameter, name: str, typ: tp.Type) -> bool:
             if param.name != name:
                 return False
 
@@ -413,7 +487,11 @@ class _ArgsExtractor:
             elif provided(param, "_converter", cattrs.Converter):
                 use.append(self.converter)
             elif provided(param, "_register", CreateRegister):
-                use.append(self.register)
+                use.append(
+                    self.register.clone(
+                        last_type=self.want, last_meta=self.meta, skip_creator=self.creator
+                    )
+                )
             elif param.annotation in (inspect._empty, tp.Any):
                 use.append(self.meta.retrieve_one(object, param.name, default=param.default))
             else:
@@ -434,37 +512,38 @@ class CreateArgs:
 
 
 class CreatorDecorator(tp.Generic[T]):
-    func: ConvertDefinition[T]
-
     def __init__(
         self,
         register: CreateRegister,
         typ: tp.Type[T],
         assume_unchanged_converted=True,
-        return_wrapped=False,
     ):
         self.typ = typ
         self.register = register
-        self.return_wrapped = return_wrapped
         self.assume_unchanged_converted = assume_unchanged_converted
 
     def __call__(self, func: tp.Optional[ConvertDefinition[T]] = None) -> ConvertDefinition[T]:
+        wrapped, func = self.wrap(func)
+        self.register[self.typ] = wrapped
+        return func
+
+    def wrap(
+        self, func: tp.Optional[ConvertDefinition[T]] = None
+    ) -> tp.Tuple[ConvertFunction[T], ConvertDefinition[T]]:
         if func is None:
-            self.func = take_or_make
+            self.func = tp.cast(ConvertDefinition[T], take_or_make)
         else:
-            self.func = func
+            self.func = tp.cast(ConvertDefinition[T], func)
 
         if hasattr(self.func, "side_effect"):
             # Hack to deal with mock objects
             self.signature = inspect.signature(self.func.side_effect)  # type: ignore
         else:
-            self.signature = inspect.signature(self.func)
 
-        if self.return_wrapped:
-            return WrappedCreator(self.wrapped, self.func)
-        else:
-            self.register[self.typ] = WrappedCreator(self.wrapped, self.func)
-            return self.func
+            self.signature = inspect.signature(tp.cast(tp.Callable, self.func))
+
+        wrapped = WrappedCreator(self.wrapped, self.func)
+        return wrapped, self.func
 
     def wrapped(self, create_args: CreateArgs) -> T:
         want = create_args.want
@@ -548,5 +627,13 @@ class CreatorDecorator(tp.Generic[T]):
             return tp.cast(ConvertDefinitionValueAndType, self.func)(value, want)
 
         else:
-            args = _ArgsExtractor(self.signature, value, want, meta, converter, register).extract()
+            args = _ArgsExtractor(
+                signature=self.signature,
+                value=value,
+                want=want,
+                meta=meta,
+                converter=converter,
+                register=register,
+                creator=self.func,
+            ).extract()
             return self.func(*args)
